#!/bin/bash
#
# /etc/libvirt/hooks/daemon
#

# HOOKLOG="./daemon-hook.log"
# TMP_DIR="./tmp/"
# OPT_DEBUG=1
HOOKLOG="/var/log/libvirt/autonuma/daemon-hook.log"
TMP_DIR="/etc/libvirt/hooks/tmp/"

# Nothing allowed outside "/bin"
# PATH="/bin:"

do_debug () {
  if [ "${OPT_DEBUG}" ]; then
    echo $*
  fi
}

# Describe the cputune (vcpupinning)
hook_vcputune () {
  node_cpu_array=(${NODE_CPU_LST})

  echo "  <cputune>"
  for ((p = 0; p < $req_cores; p++))
  do
    let n="((p / range))"
    cpuset="${node_cpu_array[$n]}"
    echo "    <vcpupin vcpu='$p' cpuset='${cpuset=}'/>"
  done
  echo "  </cputune>"
}

# Describe the NUMA nodes
hook_vcpunuma () {
  node_dist_array=(${NUMA_NODES_DIST_ARR})

  echo "    <numa>"
  for ((n = 0; n < ${NUMA_NODES}; n++))
  do
    let r_start="($n * $range)"
    let r_end="(($n + 1) * $range) - 1"
    vcpuset="$r_start-$r_end"

    echo "      <cell id='${n}' cpus='${vcpuset}' memory='${req_memory_per_node}'>"
    echo "        <distances>"

    for ((d = 0; d < ${NUMA_NODES}; d++))
    do
      echo "          <sibling id='${d}' value='${node_dist_array[(($n*${NUMA_NODES})+$d)]}'/>"
    done
    echo "        </distances>"
    echo "      </cell>"
  done
  echo "    </numa>"
}

virt_hook () {

    let range="($req_cores / ${NUMA_NODES})"

    # collect <cpu ... prior to <clock
    XMLCPU=$(echo "${XMLIN}" | awk '/<cpu/{got=1}/<clock/{got=0;prev}got' | grep -v "</cpu>");
    cnt=$(echo "${XMLCPU}" | wc -l)
    if ((cnt == 1))
    then
        XMLCPU=$(echo "${XMLCPU}" | sed -e 's/\/>/>/');
    fi

    # generate the Oracle HOOK required XML code for <cputune> and <numa>.
    XMLCPUTUNE=$(hook_vcputune $*)
    XMLCPUNUMA=$(hook_vcpunuma $*)

    do_debug ___BEGIN___
    echo "${XMLIN}" | awk '/<os>/&&!got{got=1;next}got&&/<os>/{got=0;next}!got'

    do_debug __CPUTUNE__
    echo "${XMLCPUTUNE}"

    do_debug __OS__
    if [ "${XMLCPU}" ]; then
    echo "${XMLIN}" | awk '/<os>/{got=1}/<cpu/{got=0}got'
    echo "${XMLCPU}"
    else
    echo "${XMLIN}" | awk '/<os>/{got=1}/<clock/{got=0;prev}got'
    echo "  <cpu>"
    fi

    do_debug __NUMA__
    echo "${XMLCPUNUMA}"
    echo "  </cpu>"

    do_debug ___END___
    echo "${XMLIN}" | awk '/<clock/{got=1;prev}/<\/domain>/{got=1}got'
}

build_environment () {
  # Get NUMA_NODES = the Host NUMA nodes.
  export NUMA_NODES=$(ls -1d /sys/devices/system/node/node* | wc -l);

  # Get NUMA_NODES_DIST_LST = the HOST node distances.
  export NUMA_NODES_DIST_LST=$(cat /sys/devices/system/node/node*/distance | tr ' ' ',');
  export NUMA_NODES_DIST_ARR=$(cat /sys/devices/system/node/node*/distance);

  # Get NUMA_NODES_MEM_FREE = the HOST free memory per node.
  export NUMA_NODES_MEM_FREE=$(grep MemFree /sys/devices/system/node/node*/meminfo | awk '{ print $4 }');
  export HOST_FREE_MEMORY=$(echo ${NUMA_NODES_MEM_FREE} | tr ' ' '+');

  # Get PHYS_CORES = the Physical CPU count.
  export PHYS_CORES=$(ls -1d /sys/devices/system/cpu/cpu[0-9]* | wc -l);

  # Get NODE_CPU_LST = the Physcical CPU <-> NODE mapping.
  export NODE_CPU_LST=$(cat /sys/devices/system/node/node*/cpulist);
}

prog_name=$(basename $0)
op=$2
subop=$3
extra=$4

echo "===============================================================================" >> ${HOOKLOG}
echo "$(date) '${op}/${subop}' extra = ${extra} " >> ${HOOKLOG}
echo "===============================================================================" >> ${HOOKLOG}

if [[ ${op} == "config" ]]; then

  # Does this realm carries NUMA nodes?
  if [ -d /sys/devices/system/node -a -d /sys/devices/system/node/node1 ]
  then
    # Yes...
    # prepare script logging and process the autoNUMA arrangements.
    head ${extra} >> ${HOOKLOG}
    NAME=$(basename ${extra});
    mkdir -p ${TMP_DIR}

    echo "waiting for ${extra}" >> ${HOOKLOG}
    # Setup the execution environment and report detail.
    cat ${extra} | ( build_environment ;

      # Initialize known dimensions
      let cap_numa_nodes="${NUMA_NODES}"
      let cap_phys_cores="${PHYS_CORES}"
      let cap_host_free_memory="(${HOST_FREE_MEMORY})/1024"

      XMLIN=$(awk '/<numa>/&&!got{got=1;next}got&&/<\/numa>/{got=0;next}!got' | \
        awk '/<cputune>/&&!got{got=1;next}got&&/<\/cputune>/{got=0;next}!got');
      VCPU=$(echo "${XMLIN}" | \
        awk '/<vcpu/,/<\/vcpu/{print $0}'|grep -o '>[0-9]*<'|tr -d '[<>]');
      VMEM=$(echo "${XMLIN}" | \
        awk '/<currentMemory/,/<\/currentMemory/{print $0}'|grep -o '>[0-9]*<'|tr -d '[<>]');

      let autoNUMA="${VCPU}%$cap_numa_nodes"
      if ((autoNUMA == 0))
      then
        let req_cores="(${VCPU}/$cap_numa_nodes)*$cap_numa_nodes"
        let req_memory="(${VMEM})/1024"
        let req_memory_per_node="($req_memory/${NUMA_NODES})*1024"

        virt_hook
      else
        echo "${XMLIN}"
      fi
    ) > ${TMP_DIR}/$$.$prog_name.${NAME}
    mv -f ${TMP_DIR}/$$.$prog_name.${NAME} ${extra} >> ${HOOKLOG}
    wait

    rmdir -v ${TMP_DIR} >> ${HOOKLOG}
    echo "done waiting for ${extra}" >> ${HOOKLOG}
  fi
fi

exit 0
